"""
  Copyright (c) 2022 LG Electronics Inc.
  SPDX-License-Identifier: MIT
"""

"""
webos-emulator: webOS Emulator Launcher
"""

__author__ = """Heegoo Han"""
__email__ = 'heegoo.han@lge.com'
__version__ = '0.1.0'  # TODO: is this needed?

from webos_emulator.version import __version__
from webos_emulator.__main__ import WebosEmulator



"""
  Copyright (c) 2022-2024 LG Electronics Inc.
  SPDX-License-Identifier: MIT
"""

import logging
import subprocess
from subprocess import DEVNULL   # TODO: check Python 3.3 above
import re
import locale

# TODO: set logging level
STDIN = DEVNULL  # quiet, None for info level
hostos_encoding = locale.getpreferredencoding()

def get_stderr():
    """get stderr log level by int
    """
    if logging.root.level == logging.DEBUG:
        return None
    else:
        return DEVNULL

def get_vboxmanage(command):
    """Get the full path of vboxmanage"""
    try:
        sp = subprocess.Popen([command, '-version'], stdin=DEVNULL, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        version, error = sp.communicate()
    except:
        return None, None
    else:
        return command, version

VBOXM, VBOXVER = get_vboxmanage("VBoxManage")
if VBOXVER:
    VBOXVER = str(VBOXVER, hostos_encoding).split('\n')[0]

def is_vd_exists(name):
    """Check the given vd is exists

    Args:
        name (string): target name of vd
    """
    vdcmd = VBOXM
    command = [vdcmd] + ['list', 'vms']

    try:
        sp = subprocess.Popen(command, stdin=DEVNULL, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        result, error = sp.communicate()
    except:
        return False
    result = str(result, hostos_encoding).split('\n')
    pattern = "^\"" + name + "\""
    for i in result:
        if re.search(pattern, i):
            return True
    return False

def check_linux_guest(name):
    """Check the given vm is Linux guest

    Args:
        name (string): target name
    """
    vdcmd = VBOXM
    command = [vdcmd] + ['showvminfo', name]

    try:
        sp = subprocess.Popen(command, stdin=DEVNULL, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        result, error = sp.communicate()
    except:
        return False
    result = str(result, hostos_encoding).split('\n')
    for i in result:
        if i == "":
            break
        n = i.split(":")
        if n[0]=="Guest OS":
            n1 = n[1].strip()
            if n1 == "Other Linux (64-bit)" or n1 == "Other Linux (32-bit)":
                return True
    return False

def validate_vd_name(name, listing):
    """Validate the given vd name

    Args:
        name (string): target name of vd
        listing (boolean): list option
    """
    vdcmd = VBOXM
    command = [vdcmd] + ['list', 'vms']

    try:
        sp = subprocess.Popen(command, stdin=DEVNULL, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        result, error = sp.communicate()
        command = [vdcmd] + ['list', 'runningvms']
        sp = subprocess.Popen(command, stdin=DEVNULL, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        result2, error2 = sp.communicate()
    except:
        print("webos-emulator : Please install virtualbox and set the PATH variable in the system envrionment.")
        print("On Windows, please refer to https://www.webosose.org/docs/tools/sdk/emulator/virtualbox-emulator/emulator-user-guide/#setting-the-path-on-windows")
        return ("__VBOX_NOT_INSTALLED__", "", "", "")

    dl = []
    product = "ose"
    rname = ""
    ruuid = ""
    version = ""
    if 'vboxconfig' in str(result, hostos_encoding):
        print("webos-emulator : Please recompile the kernel module and install.")
        print(str(result, hostos_encoding))
        return ("__VBOX_DRV_KERNEL_PROBLEM_", "", "", "")
    result = str(result, hostos_encoding).split('\n')
    running_vm = ""

    if result2:
        result2 = str(result2, hostos_encoding).split('\n')
        running_vm = result2[0].split("\" ")[0][1:]

    for i in result:
        if i == "":
            break
        vm = n = i.split("\" ")
        n = vm[0][1:]
        m = vm[1].strip()[1:-1]

        if check_linux_guest(n):
            dl.append((n, m))
    if listing:
        for i,j in dl:
            if i == running_vm:
                i = i + " (running)"
            print(i)
        return ("","","")
    else:
        for i,j in dl:
            if i == name or j == name:
                rname = i
                ruuid = j
                if rname.startswith("LG webOS TV Emulator"):
                    product = "tv"
                    version = rname.split("LG webOS TV Emulator")[1]
                    if version == "":
                        version = "1.2.0"
                    else:
                        version = version[1:]
                elif rname.startswith("LG webOS SIGNAGE Emulator"):
                    product = "signage"
                    version = rname.split("LG webOS SIGNAGE Emulator ")[1]
    return (rname, ruuid, product, version)

def is_vd_running(name):
    """Check the given vd is running

    Args:
        name (string): target name of vd
    """
    vdcmd = VBOXM
    command = [vdcmd] + ['list', 'runningvms']

    try:
        sp = subprocess.Popen(command, stdin=DEVNULL, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        result, error = sp.communicate()
    except:
        print("webos-emulator : is_vd_running subprocess.Popen error")
        return False
    if error:
       print("webos-emulator : Popen error %s" % error)
       return False
    result = str(result, hostos_encoding).split('\n')
    pattern = "^\"" + name + "\""
    for i in result:
        if re.search(pattern, i):
            return True
    return False

def get_storage_name(name):
    """Check the given vd's stroage controller name

    Args:
        name (string): target name of vd
    """
    vdcmd = VBOXM
    command = [vdcmd] + ['showvminfo', name]

    try:
        sp = subprocess.Popen(command, stdin=DEVNULL, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        result, error = sp.communicate()
    except:
        print("webos-emulator : get_storage_name subprocess.Popen error")
        return ""
    if error:
       print("webos-emulator : Popen error %s" % error)
       return ""
    result = str(result, hostos_encoding).split('\n')
    for i in result:
        if i.startswith("Storage Controller Name (0):"):
            found = i.split("Storage Controller Name (0):")
            return found[1].strip()
        if i.startswith("#0:"):
            found = i.split(",")[0].split()[1]
            return found.strip()[1:-1]
    return ""

def is_safe_to_create(name): # TODO: need to rename the method name
    """Check if the emulator is running

    :param name:
        the name of emulator
    """
    logging.info("is_safe_to_create : %s" % name)
    if VBOXM == None:
        print("webos-emulator : Please install virtualbox.")
        return False

    if is_vd_exists(name) and is_vd_running(name):
        return False
    else:
        return True

def detach_image(name):
    """Detach image from the given vd

    Args:
        name (string): target name of vd
    """
    vdcmd = VBOXM
    command = [vdcmd] + ["storageattach", name, "--storagectl", name,
                         "--type", "hdd", "--medium", "emptydrive",
                         "--port", "0", "--device", "0"]

    if subprocess.call(command, stdin=STDIN, stderr=get_stderr()) != 0:
        logging.debug("detach error")
        # TODO: check if storage is attached or not using grep '(UUID: ' in showvminfo
        # return False
    return True



"""
  Copyright (c) 2022 LG Electronics Inc.
  SPDX-License-Identifier: MIT
"""
class webos_emulatorError(Exception):
    """Base webos-emulator exception that others inherit.

    """
    
class VdError(webos_emulatorError):
    """vd related exception

    """
    
class DetachError(VdError):
    """vd detach error exception

    """
    def __init__(self, vdname:str):
        """Detach Error

        Args:
            vdname (str): vd name of exception occurred
        """
        super().__init__(f"[{vdname}] could not detach image")
        self.vdname = vdname



"""
  Copyright (c) 2022-2023 LG Electronics Inc.
  SPDX-License-Identifier: MIT
"""

"""Main module."""

# TODO: make webos-emulator class

import logging
import subprocess
import os, platform
import json
from subprocess import DEVNULL
from sys import stderr # TODO: check Python 3.3 above
from webos_emulator import WebosEmulator
from webos_emulator.exceptions import DetachError
import locale

# TODO: set logging level
STDIN = DEVNULL  # quiet, None for info level
hostos_encoding = locale.getpreferredencoding()

from webos_emulator.check import detach_image, get_stderr, get_storage_name, get_vboxmanage, is_safe_to_create, is_vd_exists, is_vd_running
from webos_emulator.check import VBOXM

here = os.path.abspath(os.path.dirname(__file__))
VD_JSON = json.loads(open(os.path.join(here, "webos-emulator.json"), encoding=hostos_encoding).read())

def detach_storage(name):
    """detach the image from the vd

    Args:
        name (str): vd name
    """
    logging.info("detach_storage : %s" % name)
    vdcmd = VBOXM
    command = [vdcmd] + ['storageattach', name, '--storagectl', name, '--type',
                         'hdd', '--medium', 'emptydrive', '--port', '0', '--device', '0']
    if subprocess.call(command, stdin=STDIN, stderr=get_stderr()) == 0:
        return True
    else:
        return False

def attach_storage(vdcmd, name, vdimage):
    """attach the image to the vd

    Args:
        vdcmd (str): virtualizer command
        name (str): vd name
    """
    try:
        storage_name = get_storage_name(name)
        if storage_name == "":
            storage_name = name
        command = [vdcmd] + ['storageattach', name, '--storagectl', storage_name, '--type',
                             'hdd', '--port', '0', '--device', '0', '--medium', vdimage]
        subprocess.check_call(command, stdin=STDIN, stdout=DEVNULL, stderr=get_stderr())
    except subprocess.CalledProcessError as e:
        print("webos-emulator : attach_stroage error")
        logging.debug("attach_storage error : %s" % e)
        return False
    return True

def remove_vd(name):
    """remove vd

    Args:
        name (str): vd name
    """
    logging.info("remove_vd : %s" % name)
    if VBOXM == None:
        print("webos-emulator : Please install virtualbox.")
        return False
    
    if is_vd_exists(name):
        detach_storage(name)
        command = [VBOXM] + ['unregistervm', name, '--delete']
        if subprocess.call(command, stdin=STDIN, stderr=get_stderr()) == 0:
            return True
        else:
            raise DetachError(vdname=name)
    return False

def modify_vd(vd: WebosEmulator, mstr, newname, vmdk):
    """modify vd settings

    Without extra options, it shows current settings

    Args:
        vd (WebosEmulator): vd object
        mstr : modifications settings list
        newname : new name of vm
        vmdk : vmdk file
    """
    if is_vd_running(vd.name):
        print("webos-emulator : vd is running. please stop vd before modify")
        return False
    storage_name = get_storage_name(vd.name)
    storage = storage_name + " (0, 0)"
    settings = ["Memory size", "Monitor count", "Number of CPUs", "VRAM size", "Name", "Guest OS", storage]

    tname = vd.name
    if newname:
        tname = newname

    if mstr == "" and vmdk == "":
        print("-m options:")
        print("  --memory <memory size in MB>")
        print("  --vram <video memory size in MB>")
        print("  --cpus <number>")
        print("  --monitorcount <number>")
        print("  --name <name>")
        print("  --ostype <Linux or Linux_64>")
        print("  --vmdk <vmdk file>")
        print()
    else:
        try:
            if mstr:
                command = [VBOXM] + ['modifyvm', vd.name] + mstr.split(":")[:-1]
                subprocess.check_call(command, stdin=STDIN, stdout=DEVNULL, stderr=get_stderr())
            if vmdk != "":
                command = [VBOXM] + ['storageattach', tname, '--storagectl', storage_name, '--type',
                        'hdd', '--port', '0', '--device', '0', '--medium', vmdk]
                subprocess.check_call(command, stdin=STDIN, stdout=DEVNULL, stderr=get_stderr())
        except subprocess.CalledProcessError as e:
            print("webos-emulator : modify_vd, set subprocess.Popen error")
            logging.debug("modify error : %s" % e)
            return False
    command = [VBOXM] + ['showvminfo', tname]
    try:
        sp = subprocess.Popen(command, stdin=DEVNULL, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        vdinfo, error = sp.communicate()
    except:
        print("webos-emulator : modify_vd, get subprocess.Popen error")
        return False
    if error:
       print("webos-emulator : Popen error %s" % error)
       return False
    vdinfo = str(vdinfo, hostos_encoding).split('\n')

    print("following is the current settings of vd")
    for i in vdinfo:
        if i.split(":")[0] in settings:
            print(i)
    return True

def create_vd(vd: WebosEmulator):
    """create a vd
    
    Temporal method for create vd.
    webos-emulator class will be made sooner.

    Args:
        vd (WebosEmulator): vd object
    """
    name = vd.name
    if is_safe_to_create(name):
        logging.info("%s is safe to create." % name)
        try:
            remove_vd(name)
        except DetachError as e:
            logging.error("webos-emulator error : %s" % e)
            return False
    else:
        # TODO: error handling
        return False
    
    # TODO: error handling
    try:
        logging.info("creating vd....")
        command = [VBOXM] + ['createvm', '--ostype', 'Linux_64', '--register', '--name', name]
        subprocess.check_call(command, stdin=STDIN)
        command = [VBOXM] + ['storagectl', name, '--add', 'ide', '--name', name]
        subprocess.check_call(command, stdin=STDIN)
        command = [VBOXM] + ['modifyvm', name, '--boot1', 'disk', '--boot2', 'none', '--boot3', 'none', '--boot4', 'none']
        subprocess.check_call(command, stdin=STDIN)
        
        # setting memory, videoram and cpu
        command = [VBOXM] + ['modifyvm', name, '--memory', vd.ram, '--vram', vd.vram, '--ioapic', 'on', '--cpus', vd.cpus]
        subprocess.check_call(command, stdin=STDIN)
        
        # setting graphics stuffs
        command = [VBOXM] + ['modifyvm', name, '--graphicscontroller', 'vmsvga']
        subprocess.check_call(command, stdin=STDIN)
        command = [VBOXM] + ['modifyvm', name, '--accelerate3d', 'on']
        subprocess.check_call(command, stdin=STDIN)
        
        # set usb tablet and sound
        if platform.system() == 'Windows':
            command = [VBOXM] + ['modifyvm', name, '--mouse', 'usbtablet', '--audio', 'dsound', '--audioout', 'on', '--audioin', 'on']
        elif platform.system() == 'Darwin':
            command = [VBOXM] + ['modifyvm', name, '--mouse', 'usbtablet', '--audio', 'coreaudio', '--audioout', 'on', '--audioin', 'on']
        else:
            command = [VBOXM] + ['modifyvm', name, '--mouse', 'usbtablet', '--audio', 'pulse', '--audioout', 'on', '--audioin', 'on']
        subprocess.check_call(command, stdin=STDIN)
        
        # setting network
        command = [VBOXM] + ['modifyvm', name, '--nic1', 'nat', '--natpf1', 'ssh,tcp,,'+ vd.hostssh  +',,22']
        subprocess.check_call(command, stdin=STDIN)
        command = [VBOXM] + ['modifyvm', name, '--natpf1', 'web-inspector,tcp,,9998,,9998']
        subprocess.check_call(command, stdin=STDIN)
        command = [VBOXM] + ['modifyvm', name, '--natpf1', 'enact-browser-web-inspector,tcp,,9223,,9999']
        subprocess.check_call(command, stdin=STDIN)
        
        # serial to null
        if platform.system() == 'Windows':
            command = [VBOXM] + ['modifyvm', name, '--uart1', '0x3f8', '4', '--uartmode1', 'file', 'null']
        else:
            command = [VBOXM] + ['modifyvm', name, '--uart1', '0x3f8', '4', '--uartmode1', 'file', '/dev/null']
        subprocess.check_call(command, stdin=STDIN)
        
        # two display
        command = [VBOXM] + ['modifyvm', name, '--monitorcount', '2']
        subprocess.check_call(command, stdin=STDIN)
        
        # scale factor to 0.7
        command = [VBOXM] + ['setextradata', name, 'GUI/ScaleFactor', '0.7']
        subprocess.check_call(command, stdin=STDIN)

        # set signature
        command = [VBOXM] + ['setextradata', name, 'wemul', 'ose']
        subprocess.check_call(command, stdin=STDIN)

    except subprocess.CalledProcessError as e:
        print("webos-emulator : creation error")
        logging.debug("creation error : %s" % e)
        return False
    else: # creation success
        if vd.image: # TODO: just create a vd without an image?
            if attach_storage(VBOXM, name, vd.image) == False:
                print("webos-emulator : The vmdk file is already attached. Please use a new vmdk")
                remove_vd(name)
                return False
    return True
        
def start_vd(vd: WebosEmulator):
    """start a vd

    Args:
        vd (WebosEmulator): vd object
    """
    if is_vd_exists(vd.name):
        # TODO: use is_vd_running below
        if is_safe_to_create(vd.name): # TODO: check image is attached meaning for now, we must create a vd with -i option
            if vd.product == "tv":
                # signage : LG_WEBOS_SIGNAGE_SDK_HOME + /Emulator/v4.1.7/LG_webOS_SIGNAGE_Emulator.sh
                if 'LG_WEBOS_TV_SDK_HOME' in os.environ:
                    if platform.system() == 'Windows':
                        command = os.environ['LG_WEBOS_TV_SDK_HOME'] + '\\Emulator\\v' + vd.version + '\\LG_webOS_TV_Emulator.exe'
                    elif platform.system() == 'Darwin':
                        command = 'open ' + os.environ['LG_WEBOS_TV_SDK_HOME'] + '/Emulator/v' + vd.version + '/LG_webOS_TV_Emulator_RCU.app &'
                    else:
                        command = os.environ['LG_WEBOS_TV_SDK_HOME'] + '/Emulator/v' + vd.version + '/LG_webOS_TV_Emulator.sh &'
                    # print("please run : chmod +x " +  os.environ['LG_WEBOS_TV_SDK_HOME'] + "/Resources/Jre/bin/*" )
                else:
                    print("webos-emulator : please check installation of TV Emulator")
                    return False
            elif vd.product == "signage":
                if 'LG_WEBOS_SIGNAGE_SDK_HOME' in os.environ:
                    command = os.environ['LG_WEBOS_SIGNAGE_SDK_HOME'] + '/Emulator/v' + vd.version + '/LG_webOS_SIGNAGE_Emulator.sh &'
                else:
                    print("webos-emulator : please check installation of SIGNAGE Emulator")
                    return False
            else:
                command = [VBOXM] + ['startvm', vd.name]
            
            if vd.product == "ose":
                if subprocess.call(command, stdin=STDIN , stdout=subprocess.DEVNULL,  stderr=get_stderr()) != 0:
                    print("webos-emulator : TV Emulator is needed")
                    return False
            else:
                if subprocess.call(command, stdin=STDIN , stdout=subprocess.DEVNULL,  shell=True, stderr=get_stderr()) != 0:
                    print("webos-emulator : start error")
                    return False
    else:
        print("webos-emulator : vd does not exist!")
        return False
    return True

def stop_vd(vd: WebosEmulator):
    """stop a vd

    Args:
        vd (WebosEmulator): vd object
    """
    if is_vd_exists(vd.name):
        if is_vd_running(vd.name):
            try:
                command = [VBOXM] + ['controlvm', vd.name, 'pause']
                subprocess.check_call(command, stdin=STDIN, stdout=DEVNULL, stderr=get_stderr())
                command = [VBOXM] + ['controlvm', vd.name, 'poweroff']
                subprocess.check_call(command, stdin=STDIN, stdout=DEVNULL, stderr=get_stderr())
            except subprocess.CalledProcessError as e:
                print("webos-emulator : stop error")
                logging.debug("power off error : %s" % e)
                return False
            return True
        else:
            print("webos-emulator : vd is not running.")
            return False
    return False

def delete_vd(vd: WebosEmulator):
    """delete a vd

    Args:
        vd (WebosEmulator): vd object
    """
    if is_vd_exists(vd.name):
        if not is_vd_running(vd.name):
            if detach_image(vd.name):
                command = [VBOXM] + ['unregistervm', vd.name, '--delete']
                if subprocess.call(command, stdin=STDIN, stderr=get_stderr()) != 0:
                    logging.error("webos-emulator error : delete_vd failed")
                    logging.debug("reason : unregistervm failed ")
            else:
                print("webos-emulator : delete_vd failed")
        else:
            print("webos-emulator : vd is running. please stop vd before delete")

def set_default(vd: WebosEmulator, extended):
    """set default values

    Args:
        vd (WebosEmulator): vd object
        exteneded : function extended for vs code extension integration
    """
    name = vd.name
    try:
        logging.info("set default....")
        #command = [VBOXM] + ['modifyvm', name, '--ostype', 'Linux_64']
        #subprocess.check_call(command, stdin=STDIN)

        if extended:
            command = [VBOXM] + ['storagectl', name, '--add', 'ide', '--name', name]
            subprocess.check_call(command, stdin=STDIN)
            command = [VBOXM] + ['modifyvm', name, '--boot1', 'disk', '--boot2', 'none', '--boot3', 'none', '--boot4', 'none']
            subprocess.check_call(command, stdin=STDIN)
            if vd.vmdkfile:
                command = [VBOXM] + ['storageattach', name, '--storagectl', name, '--type',
                         'hdd', '--port', '0', '--device', '0', '--medium', vd.vmdkfile]
                subprocess.check_call(command, stdin=STDIN)

        # setting memory, videoram and cpu
        command = [VBOXM] + ['modifyvm', name, '--memory', vd.ram, '--vram', vd.vram, '--ioapic', 'on', '--cpus', vd.cpus]
        subprocess.check_call(command, stdin=STDIN)

        # setting graphics stuffs
        command = [VBOXM] + ['modifyvm', name, '--graphicscontroller', 'vmsvga']
        subprocess.check_call(command, stdin=STDIN)
        command = [VBOXM] + ['modifyvm', name, '--accelerate3d', 'on']
        subprocess.check_call(command, stdin=STDIN)

        # set usb tablet and sound
        if platform.system() == 'Windows':
            command = [VBOXM] + ['modifyvm', name, '--mouse', 'usbtablet', '--audio', 'dsound', '--audioout', 'on', '--audioin', 'on']
        elif platform.system() == 'Darwin':
            command = [VBOXM] + ['modifyvm', name, '--mouse', 'usbtablet', '--audio', 'coreaudio', '--audioout', 'on', '--audioin', 'on']
        else:
            command = [VBOXM] + ['modifyvm', name, '--mouse', 'usbtablet', '--audio', 'pulse', '--audioout', 'on', '--audioin', 'on']
        subprocess.check_call(command, stdin=STDIN)

        # remove network
        #sp = subprocess.Popen([command, '-version'], stdin=DEVNULL, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        #version, error = sp.communicate()

        try:
            command = [VBOXM] + ['modifyvm', name, '--nic1', 'nat', '--natpf1', 'delete', 'ssh']
            subprocess.check_call(command, stdin=STDIN, stdout=DEVNULL, stderr=get_stderr()) # virtualbox7 needs "stderr=get_stderr()"
        except:
            pass
        try:
            command = [VBOXM] + ['modifyvm', name, '--nic1', 'nat', '--natpf1', 'delete', 'web-inspector']
            subprocess.check_call(command, stdin=STDIN, stdout=DEVNULL, stderr=get_stderr())
        except:
            pass
        try:
            command = [VBOXM] + ['modifyvm', name, '--nic1', 'nat', '--natpf1', 'delete', 'enact-browser-web-inspector']
            subprocess.check_call(command, stdin=STDIN, stdout=DEVNULL, stderr=get_stderr())
        except:
            pass

        # setting network
        command = [VBOXM] + ['modifyvm', name, '--nic1', 'nat', '--natpf1', 'ssh,tcp,,'+ vd.hostssh  +',,22']
        subprocess.check_call(command, stdin=STDIN)
        command = [VBOXM] + ['modifyvm', name, '--natpf1', 'web-inspector,tcp,,9998,,9998']
        subprocess.check_call(command, stdin=STDIN)
        command = [VBOXM] + ['modifyvm', name, '--natpf1', 'enact-browser-web-inspector,tcp,,9223,,9999']
        subprocess.check_call(command, stdin=STDIN)

        # serial to null
        if platform.system() == 'Windows':
            command = [VBOXM] + ['modifyvm', name, '--uart1', '0x3f8', '4', '--uartmode1', 'file', 'null']
        else:
            command = [VBOXM] + ['modifyvm', name, '--uart1', '0x3f8', '4', '--uartmode1', 'file', '/dev/null']
        subprocess.check_call(command, stdin=STDIN)

        # monitor
        if extended:
            command = [VBOXM] + ['modifyvm', name, '--monitorcount', vd.monitorcount]
        else:
            command = [VBOXM] + ['modifyvm', name, '--monitorcount', '2']
        subprocess.check_call(command, stdin=STDIN)

        # scale factor to 0.7
        if extended:
            command = [VBOXM] + ['setextradata', name, 'GUI/ScaleFactor', vd.scalefactor]
        else:
            command = [VBOXM] + ['setextradata', name, 'GUI/ScaleFactor', '0.7']
        subprocess.check_call(command, stdin=STDIN)

        # set signature
        command = [VBOXM] + ['setextradata', name, 'wemul', 'ose']
        subprocess.check_call(command, stdin=STDIN)

    except subprocess.CalledProcessError as e:
        print("webos-emulator : setting error")
        logging.debug("setting error : %s" % e)
        return False

    return True

def default_vd(vd: WebosEmulator):
    """set to default settings

    Args:
        vd (WebosEmulator): vd object
    """
    if is_vd_exists(vd.name):
        if not is_vd_running(vd.name):
            if set_default(vd, False):
                return True
            else:
                print("webos-emulator : default_vd failed")
        else:
            print("webos-emulator : vd is running. please stop vd before setting default")
    return False

def hidden_create(vd: WebosEmulator):
    """hidden create vd

    Args:
        vd (WebosEmulator): vd object
    """
    if is_vd_exists(vd.name):
        if not is_vd_running(vd.name):
            if set_default(vd, True):
                return True
            else:
                print("webos-emulator : hidden_create failed")
        else:
            print("webos-emulator : vd is running. please stop vd before hidden create")
    return False

def custom_vd(vd: WebosEmulator, ovafile):
    """set to default settings

    Args:
        vd (WebosEmulator): vd object
        ovafile: ova format 1.0 file
    """
    if not is_vd_exists(vd.name):
        try:
            logging.info("custom vd....")
            command = [VBOXM] + ['import', ovafile, '--vsys', '0', '--vmname', vd.name]
            subprocess.check_call(command, stdin=STDIN, stdout=DEVNULL, stderr=get_stderr())
            command = [VBOXM] + ['modifyvm', vd.name, '--boot1', 'disk', '--boot2', 'none', '--boot3', 'none', '--boot4', 'none']
            subprocess.check_call(command, stdin=STDIN, stdout=DEVNULL)
            old_name = get_storage_name(vd.name)
            if old_name:
                command = [VBOXM] + ['storagectl', vd.name, '--name', old_name, '--rename', vd.name]
                subprocess.check_call(command, stdin=STDIN, stdout=DEVNULL)
        except subprocess.CalledProcessError as e:
            print("webos-emulator : custom error")
            logging.debug("custom error : %s" % e)
            return False
        else: # creation success
            if vd.image: # TODO: just create a vd without an image?
                if attach_storage(VBOXM, vd.name, vd.image): # TODOL check ok
                    print("webos-emulator : custom error")
                    return False
        return True
    else:
        print("webos-emulator : vd is exist. please delete vd before setting custom file")
    return False



"""
  Copyright (c) 2022 LG Electronics Inc.
  SPDX-License-Identifier: MIT
"""

"""
This module implements the developer interface for webos-emulator.

:class:`WebosEmulator <WebosEmulator> class focuses on the developer interface.
    
"""
from typing import Optional

import webos_emulator
class WebosEmulator:
    """Core developer interface for webos-emulator."""
    
    def __init__(
        self,
        name: str,
        vdname: str
    ):
        """Construct a :class:`WebosEmulator <WebosEmulator>`.
        
        :param str name:
            A webOS emulator name.
        """
        self._name = name
        self._product = "ose"  # TODO
        self._version = "TBD"  # TODO
        self._cpus: Optional[str] = '2'  # number of CPUs
        self._ram: Optional[str] = '4096'  # vd RAM, MBs
        self._vram: Optional[str] = '128'  # vd video RAM, MBs
        self._hostssh: Optional[str] = '6622' # host to emulator ssh port number
        self._image: Optional[str] = None # image name if exists
        self._monitorcount: Optional[str] = '2'  # number of monitors
        self._scalefactor: Optional[str] = '0.7'  # number of scale factor
        self._vmdkfile: Optional[str] = ''  # vmdkfile
        
    @property
    def name(self):
        """Return webOS emulator name"""
        return self._name

    @property
    def product(self):
        """Return webOS emulator product name"""
        return self._product

    @property
    def version(self):
        """Return webOS emulator version name"""
        return self._version

    @property
    def cpus(self):
        """Return webOS emulator number of CPUs """
        return self._cpus
        
    @property
    def ram(self):
        """Return webOS emulator RAM size in MBs"""
        return self._ram
        
    @property
    def vram(self):
        """Return webOS emulator Video RAM size in MBs"""
        return self._vram
        
    @property
    def hostssh(self):
        """Return emulator ssh port number"""
        return self._hostssh
    
    @property
    def image(self):
        """Return image"""
        return self._image

    @property
    def monitorcount(self):
        """Return webOS emulator number of monitors """
        return self._monitorcount

    @property
    def scalefactor(self):
        """Return webOS emulator number of scale factor """
        return self._scalefactor

    @property
    def vmdkfile(self):
        """Return webOS emulator full path of vmdkfile """
        return self._vmdkfile

    @property
    def version(self):
        """Return webOS emulator version """
        return self._version
    
    @image.setter
    def image(self, value):
        """Sets the image"""
        self._image = value

    @ram.setter
    def ram(self, value):
        """Sets the ram"""
        self._ram = value

    @product.setter
    def product(self, value):
        """Sets the product"""
        self._product = value

    @monitorcount.setter
    def monitorcount(self, value):
        """Sets the monitor count"""
        self._monitorcount = value

    @scalefactor.setter
    def scalefactor(self, value):
        """Sets the scale factor"""
        self._scalefactor = value

    @vmdkfile.setter
    def vmdkfile(self, value):
        """Sets the vmdkfile"""
        self._vmdkfile = value

    @version.setter
    def version(self, value):
        """Sets the version"""
        self._version = value

    def create(self):
        """Create a webOS emulator
            
        Default emulator is webOS OSE emulator.
        This method will be removed soon.
            
        """
        pass
        


"""
  Copyright (c) 2022-2024 LG Electronics Inc.
  SPDX-License-Identifier: MIT
"""

"""Console script for webos-emulator."""
import argparse
import sys
from typing import List, Optional
import logging
import os

from webos_emulator import __version__
from webos_emulator import WebosEmulator
from webos_emulator.webos_emulator import attach_storage, create_vd, custom_vd, default_vd, delete_vd, hidden_create, modify_vd, set_default, start_vd, stop_vd, VD_JSON
from webos_emulator.check import VBOXM, validate_vd_name, is_vd_exists, is_vd_running

def main():
    """webOS Emulator Launcher"""
    logging.basicConfig(format='[%(levelname)s] %(message)s')
    parser = argparse.ArgumentParser(description=main.__doc__)
    args = _parse_args(parser)
    
    if args.debug:
        logging.getLogger().setLevel(logging.DEBUG)
    
    """if args.vd_image:
        vd = WebosEmulator("webos-imagex")
        vd.image = args.vd_image
        if create_vd(vd):
            start_vd(vd)
        return 0
    """
   
    if args.express:
        vd = WebosEmulator("webos-imagex", "webos-imagex")
        if args.express != "configured":
            if is_vd_running(vd.name):
                print("If you want to launch emulator, please start emualtor as below")
                print("webos-emulator -x")
            if os.path.isfile(args.express):
                vd.image = args.express
            else:
                print("webos-emulator : Please check %s exists." % args.express)
                return 1
            if VD_JSON['ram']:
                vd.ram = VD_JSON['ram']
            if create_vd(vd) == False:  # TODO: create webos-emulator class and use
                print("webos-emulator : failed")
                return 1

        vd.product = "ose"
        if is_vd_exists(vd.name):
            if is_vd_running(vd.name):
                stop_vd(vd)
            else:
                start_vd(vd)
        else:
            print("Please specify a vmdk full path to make and launch emulator like below")
            print("   webos-emulator -x  /path/to/abc.vmdk")
            print("If you have made the emulator, you can launch or kill the emulator as")
            print("   webos-emulator -x")
            return 1
        return 0
    
    if args.list:
        validate_vd_name("__list_images__", True)
        return 0
    
    name = ""
    uuid = ""
    product = "ose"
    name,uuid,product,version = validate_vd_name(args.vd, False)
    if name == "__VBOX_NOT_INSTALLED__":
        return 1
    if name == "__VBOX_DRV_KERNEL_PROBLEM_":
        return 1

    if args.vd:
        if product == "tv" or product == "signage":
            if args.modify or args.hidden_create or args.delete or args.image or args.default:
                print("Only start and kill commands are permitted for TV/Signage Emulator.")
                return 1

    if args.create:
        if args.vd == None:
            print("Please specify a vd name with -vd <name>")
            return 1
        #if not args.vd.startswith("ose_"):
        #    print("Please specify vd name as ose_version like ose_475")
        #    return 2
        ose_name = args.vd # TODO: need to check args.vd is product_version
        vd = WebosEmulator(ose_name, args.vd)
        if args.image:
            if os.path.isfile(args.image):
                vd.image = args.image
            else:
                print("webos-emulator : Please check %s exists." % args.image)
                return 1
        if VD_JSON['ram']:
            vd.ram = VD_JSON['ram']
        create_vd(vd)  # TODO: create webos-emulator class and use
        return 0

    if args.custom:
        if args.vd == None:
            print("Please specify a vd name with -vd <name>")
            return 1
        #if not args.vd.startswith("ose_"):
        #    print("Please specify vd name as ose_version like ose_475")
        #    return 2
        ose_name = args.vd # TODO: need to check args.vd is product_version
        vd = WebosEmulator(ose_name, args.vd)
        if args.image:
            if os.path.isfile(args.image):
                vd.image = args.image
            else:
                print("webos-emulator : Please check %s exists." % args.image)
                return 1
        if VD_JSON['ram']:
            vd.ram = VD_JSON['ram']
        if custom_vd(vd, args.custom) == False:  # TODO: create webos-emulator class and use
            return 1
        return 0

    if args.modify or args.hidden_create or args.start or args.stop or args.delete or args.default:
        if args.vd == None:
            print("Please specify a vd name with -vd <name>")
            return 1
        if name == "":
            print("Please check vd list via webos-emulator -l")
            return 1

    if args.modify:
        mstr = ""
        if args.memory:
            mstr = mstr + "--memory:" + str(args.memory) +":"
        if args.vram:
            mstr = mstr + "--vram:" + str(args.vram)+":"
        if args.cpus:
            mstr = mstr + "--cpus:" + str(args.cpus)+":"
        if args.monitorcount:
            mstr = mstr + "--monitorcount:" + str(args.monitorcount)+":"
        if args.name:
            mstr = mstr + "--name:" + str(args.name)+":"
        if args.ostype:
            if args.ostype != "Linux" and args.ostype != "Linux_64":
                print("Please specify a correct ostype name: Linux or Linux_64")
                return 1
            mstr = mstr + "--ostype:" + str(args.ostype)+":"
        vd = WebosEmulator(name, args.vd)
        vmdk = ""
        if args.vmdk:
            vmdk = args.vmdk.name
        modify_vd(vd, mstr, args.name, vmdk)
    elif args.hidden_create:
        vd = WebosEmulator(name, args.vd)
        mstr = ""
        if args.memory:
            vd.ram = str(args.memory)
        if args.vram:
            vd.vram = str(args.vram)
        if args.cpus:
            vd.cpus = str(args.cpus)
        if args.monitorcount:
            vd.monitorcount = str(args.monitorcount)
        if args.scalefactor:
            vd.scalefactor = str(args.scalefactor)
        if args.name:
            vd.name = str(args.name)
        if args.ostype:
            if args.ostype != "Linux" and args.ostype != "Linux_64":
                print("Please specify a correct ostype name: Linux or Linux_64")
                return 1
            vd.name = str(args.ostype)
        vmdk = ""
        if args.vmdk:
            vd.vmdkfile = str(args.vmdk.name)
        hidden_create(vd)
    elif args.image:
        if not os.path.isfile(args.image):
            print("webos-emulator : Please check %s exists." % args.image)
            return 1
        if name == "":
            print("Please specify a existing vd name")
            return 1
        vd = WebosEmulator(name, args.vd)
        vd.image = args.image
        if VD_JSON['ram']:
            vd.ram = VD_JSON['ram']
        # create_vd(vd)  # TODO: create webos-emulator class and use
        attach_storage(VBOXM, vd.name, vd.image)
    elif args.start:
        vd = WebosEmulator(name, args.vd)
        if product == "tv":
            vd.product = "tv"
            vd.version = version
        elif product == "signage":
            vd.product = "signage"
            vd.version = version
        start_vd(vd)
    elif args.stop:
        vd = WebosEmulator(name, args.vd)
        stop_vd(vd)
    elif args.delete:
        vd = WebosEmulator(name, args.vd)
        delete_vd(vd)
    elif args.default:
        vd = WebosEmulator(name, args.vd)
        default_vd(vd)
    elif args.custom:
        vd = WebosEmulator(name, args.vd)
        custom_vd(vd, args.custom)
    else:
        parser.print_help()

    return 0

def _parse_args(parser: argparse.ArgumentParser, args: Optional[List] = None) -> argparse.Namespace:
    vd_grp = parser.add_argument_group('Commands')
    vd_grp = vd_grp.add_mutually_exclusive_group()
    vd_grp.add_argument(
        "-l",
        "--list",
        action="store_true",
        dest="list",
        help="list all the vd names",
    )
    parser.add_argument(
        "-vd",  # change vm option to vd
        metavar='<name> or <uuid>',
        help="use a webOS emulator name as product_version like ose_475",
    )
    vd_grp.add_argument(
        "-m",
        "--modify",
        action="store_true",
        dest="modify",
        help="Modify a webOS emulator settings",
    )
    vd_grp.add_argument(
        "--hidden-vce-create",
        action="store_true",
        dest="hidden_create",
        help=argparse.SUPPRESS,
    )

    parser.add_argument("--memory", type=int, help=argparse.SUPPRESS) # https://docs.python.org/3/library/argparse.html
    parser.add_argument("--vram", type=int, help=argparse.SUPPRESS)
    parser.add_argument("--cpus", type=int, help=argparse.SUPPRESS)
    parser.add_argument("--monitorcount", type=int, help=argparse.SUPPRESS)
    parser.add_argument("--name", type=str, help=argparse.SUPPRESS)
    parser.add_argument("--ostype", type=str, help=argparse.SUPPRESS)
    parser.add_argument("--vmdk", type=argparse.FileType('r'), help=argparse.SUPPRESS)
    parser.add_argument("--scalefactor", type=str, help=argparse.SUPPRESS)

    #parser.add_argument(
    #    "vd_image", help="create and run vd_image (ose emulator only)", nargs="?"
    #)
    parser.add_argument(
        "--version", action="version", version="%(prog)s " + __version__,
    )
    vd_grp.add_argument(
        "-c",
        "--create",
        action="store_true",
        dest="create",
        help="Create a webOS emulator",
    )
    parser.add_argument(
        "-i",
        "--image",
        metavar='<file>',
        help="specify virtualbox image file",
    )
    vd_grp.add_argument(
        "-s",
        "--start",
        action="store_true",
        dest="start",
        help="Start a webOS emulator",
    )
    vd_grp.add_argument(
        "-k",
        "--kill",
        action="store_true",
        dest="stop",
        help="Kill a running webOS emulator",
    )
    vd_grp.add_argument(
        "-d",
        "--delete",
        action="store_true",
        dest="delete",
        help="Delete a webOS emulator",
    )
    vd_grp.add_argument(
        "-ds",
        "--default-settings",
        action="store_true",
        dest="default",
        help="Set to default settings",
    )
    vd_grp.add_argument(
        "-cc",
        "--create-with-custom",
        action="store",
        metavar='<.ova>',
        dest="custom",
        help="Create a emulator with a custom settings using OVF 1.0 file",
    )
    vd_grp.add_argument(
        "-x",
        "--express",
        nargs='?',
        action="store",
        const="configured",
        metavar='<.vmdk>',
        dest="express",
        help='Launch a emulator if vmdk is given, without vmdk option launch or kill the emulator',
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        dest="debug",
        help="Show debug info",
    )
        
    return parser.parse_args(args)

if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



#!/usr/bin/env python

"""
  Copyright (c) 2022 LG Electronics Inc.
  SPDX-License-Identifier: MIT
"""

"""The setup script."""
import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

with open(os.path.join(here, "webos_emulator", "version.py")) as version_file:
    exec(version_file.read())

requirements = [ ]

test_requirements = ['pytest>=3', ]

setup(
    author="Heegoo Han",
    author_email='heegoo.han@lge.com',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="webOS Emulator Launcher",
    entry_points={
        'console_scripts': [
            'webos-emulator=webos_emulator.cli:main',
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='webos-emulator',
    name='webos-emulator',
    packages=find_packages(include=['webos_emulator', 'webos-emulator.*']),
    package_data={'webos_emulator': ['webos-emulator.json']},
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/webosose/webos-emulator',
    version=__version__,
    zip_safe=False,
)
